# -*- coding: utf-8 -*-

import os
import re
import json
from copy import deepcopy
from typing import Dict, Optional, List, Tuple

from jinja2 import Template

from ..config import Tasks, Languages, TEMPLATE_DIR, NERFormatter, ClassificationFormatter
from ..models import GPTModel
from .base import BaseAnnotator


class GPTAnnotator(BaseAnnotator):
    def __init__(self,
                 model: GPTModel,
                 task: Tasks,
                 label_mapping: Dict,
                 language: Languages,
                 **kwargs) -> None:
        super().__init__()
        self.model = model
        self.task = task
        self.label_mapping = label_mapping
        # load template
        with open(os.path.join(TEMPLATE_DIR, 'texts', f'{task}.{language}.jinja')) as reader:
            self.template = Template(reader.read())

    @staticmethod
    def get_all_ner_segments(text: str, entities: List[str]) -> List[Tuple]:
        prev_start = 0
        segments = []
        for start, end, entity, entity_type in entities:
            if start > prev_start:
                segments.append((text[prev_start: start], 'O'))
            segments.append((entity, entity_type))
            prev_start = end
        if prev_start < len(text):
            segments.append((text[prev_start:], 'O'))
        return segments

    @staticmethod
    def format_ner_to_bio(text: str, entities: List[str]) -> str:
        segments = GPTAnnotator.get_all_ner_segments(text, entities)
        ret = []
        for segment, label in segments:
            if label == 'O':
                ret.append('\n'.join([f'{w}\t{t}' for w, t in zip(segment, [label] * len(segment))]))
            else:
                head = f'{segment[0]}\tB-{label}'
                if len(segment) > 1:
                    head += '\n'
                    head += '\n'.join([f'{w}\tI-{t}' for w, t in zip(segment[1:], [label] * len(segment[1:]))])
                ret.append(head)
        return '\n'.join(ret)

    @staticmethod
    def format_ner_to_segment(text: str, entities: List[str]) -> str:
        segments = GPTAnnotator.get_all_ner_segments(text, entities)
        ret = []
        for segment, label in segments:
            ret.append(f'{segment}\t{label}')
        return '\n'.join(ret)

    @staticmethod
    def make_ner_extraction_regex(labels: List[str]):
        # sort by label length to support overlap labels
        labels = sorted(labels, key=lambda x: len(x), reverse=True)
        return re.compile(
            r'\((?P<entity>((?!\)).)+)\,'  # entity
            r'.*?(?P<entity_type>(%s)).*?\)' % '|'.join(  # entity_type
                [re.escape(k) for k in labels]))

    @staticmethod
    def make_classification_extraction_regex(labels: List[str]):
        # sort by label length to support overlap labels
        labels = sorted(labels, key=lambda x: len(x), reverse=True)
        return re.compile(
            r'(%s)+' % '|'.join([re.escape(label) for label in labels])
        )

    def ner(self, text: str, hint: Optional[str] = None, formatter: Optional[NERFormatter] = None):
        if formatter is not None and formatter not in NERFormatter.values():
            raise ValueError(
                f'Invalid formatter `{formatter}`, please specify formatter from NERFormatter for the NER task.')
        prompt = self.template.render(
            labels=list(self.label_mapping.keys()),
            text=text,
            hint=hint)
        resp = self.model.predict(prompt)
        ret = deepcopy(resp)
        ret['result'] = {}
        ret['result']['text'] = text
        entity_map = {}
        pair_regex = GPTAnnotator.make_ner_extraction_regex(self.label_mapping.keys())
        for match in pair_regex.finditer(ret['response']):
            entity, entity_type = match.group('entity'), match.group('entity_type')
            entity = re.sub(r'(^[\'"\s]+|[\'"\s]+$)', '', entity)
            entity_type = re.sub(r'(^[\'"\s]+|[\'"\s]+$)', '', entity_type)
            if entity_type not in self.label_mapping:
                continue
            if entity not in text:
                continue
            entity_map[entity] = self.label_mapping[entity_type]
        if not entity_map:
            return ret
        # find positions
        safe_entities = [re.escape(ent) for ent in entity_map]
        entities = []
        for matched in re.finditer(r'(%s)' % ('|'.join(safe_entities)), text):
            entity = matched.group()
            start, end = matched.span()
            entities.append((start, end, entity, entity_map[entity]))
        ret['result']['entities'] = entities
        if formatter is not None:
            if formatter == NERFormatter.BIO:
                ret['result']['formatted_result'] = GPTAnnotator.format_ner_to_bio(
                    ret['result']['text'], ret['result']['entities'])
            elif formatter == NERFormatter.Segment:
                ret['result']['formatted_result'] = GPTAnnotator.format_ner_to_segment(
                    ret['result']['text'], ret['result']['entities'])
        return ret

    def classification(self,
                       text: str,
                       hint: Optional[str] = None,
                       formatter: Optional[ClassificationFormatter] = None):
        if formatter is not None and formatter not in ClassificationFormatter.values():
            raise ValueError(
                f'Invalid formatter `{formatter}`, '
                'please specify formatter from ClassificationFormatter for the NER task.')
        prompt = self.template.render(
            labels=list(self.label_mapping.keys()),
            text=text,
            hint=hint)
        resp = self.model.predict(prompt)
        ret = deepcopy(resp)
        ret['result'] = {}
        ret['result']['text'] = text
        label_regex = self.make_classification_extraction_regex(list(self.label_mapping.keys()))
        labels = label_regex.findall(ret['response'])
        if not labels:
            return ret
        label = labels[0]
        ret['result']['label'] = label
        if formatter is not None:
            if formatter == ClassificationFormatter.JSONL:
                ret['result']['formatted_result'] = json.dumps(
                    {'text': ret['result']['text'], 'label': ret['result']['label']}, ensure_ascii=False)
        return ret

    def tag(self, text: str, hint: Optional[str] = None, formatter=None):
        text = text.replace('\n', '')
        if self.task == Tasks.NER:
            return self.ner(text, hint=hint, formatter=formatter)
        elif self.task == Tasks.Classification:
            return self.classification(text, hint=hint, formatter=formatter)

    def __call__(self, text: str, hint: Optional[str] = None, formatter=None):
        return self.tag(text, hint=hint, formatter=formatter)



# -*- coding: utf-8 -*-

import time
import functools


def with_taken_time(func):
    ''' Append taken time to data
    '''
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        data = func(*args, **kwargs)
        data['meta']['taken_time'] = round(time.time() - start_time, 5)
        return data
    return wrapper



# -*- coding: utf-8 -*-

from typing import Dict, Union
from functools import partial, lru_cache

import tiktoken
import openai

from .base import BaseModel
from ..config import OpenAIModels, MAX_LRU_CACHE_SIZE
from ..utils import with_taken_time


class GPTModel(BaseModel):
    def __init__(self,
                 api_key: str,
                 model: OpenAIModels = OpenAIModels.ChatGPT,
                 temperature: float = 0.8,
                 max_tokens: int = 4000,
                 top_p: float = 1,
                 frequency_penalty: float = 0,
                 presence_penalty: float = 0,
                 stop: Union[str, None] = None,) -> None:
        super().__init__()
        self.tokenizer = tiktoken.encoding_for_model(model)
        self.model = model
        # init openai
        self.max_tokens = max_tokens
        openai.api_key = api_key
        if self.model == OpenAIModels.ChatGPT:
            self._predict = partial(
                openai.ChatCompletion.create,
                model=self.model,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                stop=stop
            )
        else:
            self._predict = partial(
                openai.Completion.create,
                model=self.model,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                stop=stop
            )

    @lru_cache(maxsize=MAX_LRU_CACHE_SIZE)
    @with_taken_time
    def predict(self, text: str) -> Dict:
        tokens = self.tokenizer.encode(text)
        if len(tokens) > self.max_tokens:
            raise ValueError(
                f'OOT (Out Of Tokens)! the current input text has `{len(tokens)}` tokens, '
                'which is greater than the max_tokens {self.max_tokens}')
        data = self.get_output_template()
        data['request'] = {'prompt': text}
        meta = {}
        if self.model == OpenAIModels.ChatGPT:
            resp = self._predict(messages=[{"role": "user", "content": text}],
                                 max_tokens=self.max_tokens - len(tokens))
            data["response"] = resp["choices"][0]["message"]["content"]
            meta["role"] = resp["choices"][0]["message"]["role"]
        else:
            resp = self._predict(prompt=text, max_tokens=self.max_tokens - len(tokens))
            data["response"] = resp["choices"][0]["text"]
        meta.update(resp["usage"])
        data['meta'] = meta
        return data



# -*- coding: utf-8 -*-

import os
from abc import ABCMeta


PACKAGE_DIR = os.path.dirname(os.path.realpath(__file__))
TEMPLATE_DIR = os.path.join(PACKAGE_DIR, "templates")
MAX_LRU_CACHE_SIZE = 100000


class AttributeClass(ABCMeta):
    @classmethod
    def list_attributes(cls):
        return [k for k in cls.__dict__.keys()
                if not k.startswith('__') and not k.endswith('__') and k not in ['list_attributes']]

    @classmethod
    def values(cls):
        return [v for k, v in cls.__dict__.items()
                if not k.startswith('__') and not k.endswith('__') and k not in ['list_attributes']]


class OpenAIModels(AttributeClass):
    '''
    OpenAI API: https://platform.openai.com/docs/models/gpt-3-5
    '''
    ChatGPT = 'gpt-3.5-turbo'
    TextDavinci003 = 'text-davinci-003'
    TextDavinci002 = 'text-davinci-002'


class Tasks(AttributeClass):
    ''' Supported Tasks
    '''
    NER = 'ner'
    Classification = 'classification'


class Languages(AttributeClass):
    ''' Supported Languages
    '''
    ZH_CN = 'zh_cn'
    EN = 'en'


class NERFormatter(AttributeClass):
    BIO = 'BIO'
    Segment = 'segment'


class ClassificationFormatter(AttributeClass):
    JSONL = 'jsonl'



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from typing import Dict


class BaseModel(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.model = None

    @abstractmethod
    def predict(self):
        raise NotImplementedError

    def get_output_template(self) -> Dict:
        return {
            'request': None,
            'meta': None,
            'response': None,
            'result': None
        }



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod


class BaseAnnotator(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.model = None

    @abstractmethod
    def tag(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#


from setuptools import setup, find_packages


__version__ = '0.1.4'


long_description = open('README.md', encoding='utf-8').read()

with open('requirements.txt', encoding='utf-8') as f:
    requirements = [l for l in f.read().splitlines() if l]

with open('dev-requirements.txt', encoding='utf-8') as f:
    test_requirements = [l for l in f.read().splitlines() if l][1:]


setup(
    name='lanno',
    version=__version__,
    description='Let Large Language Models Serve As Data Annotators.',
    long_description_content_type="text/markdown",
    long_description=long_description,
    keywords='LLM,NLP,Annotator,Prompt',
    author='sean lee',
    author_email='xmlee97@gmail.com',
    license='Apache 2.0 License',
    platforms=['all'],
    url='https://github.com/SeanLee97/lanno',
    packages=find_packages(exclude=('test*', )),
    package_data={
        '': ['*.txt', '*.jinja'],
        'lanno': ['templates/texts/*.jinja']
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',
    ],
    install_requires=requirements,
    tests_require=test_requirements,
)
